# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import operator
import os
import time
import uuid

from keystoneauth1 import discover

import openstack.config
from openstack import connection
from openstack.tests import base


#: Defines the OpenStack Client Config (OCC) cloud key in your OCC config
#: file, typically in $HOME/.config/openstack/clouds.yaml. That configuration
#: will determine where the functional tests will be run and what resource
#: defaults will be used to run the functional tests.
TEST_CONFIG = openstack.config.OpenStackConfig()
TEST_CLOUD_NAME = os.getenv('OS_CLOUD', 'devstack-admin')
TEST_CLOUD_REGION = openstack.config.get_cloud_region(cloud=TEST_CLOUD_NAME)


def _get_resource_value(resource_key):
    return TEST_CONFIG.get_extra_config('functional').get(resource_key)


def _disable_keep_alive(conn):
    sess = conn.config.get_session()
    sess.keep_alive = False


class BaseFunctionalTest(base.TestCase):
    user_cloud: connection.Connection
    user_cloud_alt: connection.Connection
    operator_cloud: connection.Connection

    _wait_for_timeout_key = ''

    def setUp(self):
        super().setUp()

        self._system_admin_name = os.environ.get(
            'OPENSTACKSDK_SYSTEM_ADMIN_CLOUD',
            'devstack-system-admin',
        )
        if not self._system_admin_name:
            raise self.failureException(
                "OPENSTACKSDK_SYSTEM_ADMIN_CLOUD must be set to a non-empty "
                "value"
            )

        self.config = openstack.config.OpenStackConfig()

        self._user_cloud_name = os.environ.get(
            'OPENSTACKSDK_DEMO_CLOUD', 'devstack'
        )
        if not self._user_cloud_name:
            raise self.failureException(
                "OPENSTACKSDK_DEMO_CLOUD must be set to a non-empty value"
            )

        self._user_alt_cloud_name = os.environ.get(
            'OPENSTACKSDK_DEMO_CLOUD_ALT', 'devstack-alt'
        )
        if not self._user_alt_cloud_name:
            raise self.failureException(
                "OPENSTACKSDK_DEMO_CLOUD_ALT must be set to a non-empty value"
            )

        self._set_user_cloud()

        self._operator_cloud_name = os.environ.get(
            'OPENSTACKSDK_OPERATOR_CLOUD', 'devstack-admin'
        )
        if not self._operator_cloud_name:
            raise self.failureException(
                "OPENSTACKSDK_OPERATOR_CLOUD must be set to a non-empty value"
            )

        self._set_operator_cloud()

        self.flavor = self._pick_flavor()
        self.image = self._pick_image()

        # Defines default timeout for wait_for methods used
        # in the functional tests
        self._wait_for_timeout = int(
            os.getenv(
                self._wait_for_timeout_key,
                os.getenv('OPENSTACKSDK_FUNC_TEST_TIMEOUT', 300),
            )
        )

    def _set_user_cloud(self, **kwargs):
        user_config = self.config.get_one(
            cloud=self._user_cloud_name, **kwargs
        )
        self.user_cloud = connection.Connection(config=user_config)
        _disable_keep_alive(self.user_cloud)

        user_config_alt = self.config.get_one(
            cloud=self._user_alt_cloud_name, **kwargs
        )
        self.user_cloud_alt = connection.Connection(config=user_config_alt)
        _disable_keep_alive(self.user_cloud_alt)

    def _set_operator_cloud(self, **kwargs):
        operator_config = self.config.get_one(
            cloud=self._operator_cloud_name, **kwargs
        )
        self.operator_cloud = connection.Connection(config=operator_config)
        _disable_keep_alive(self.operator_cloud)

        system_admin_config = self.config.get_one(
            cloud=self._system_admin_name, **kwargs
        )
        self.system_admin_cloud = connection.Connection(
            config=system_admin_config
        )
        _disable_keep_alive(self.system_admin_cloud)

    def _pick_flavor(self):
        """Pick a sensible flavor to run tests with.

        This returns None if the compute service is not present (e.g.
        ironic-only deployments).
        """
        if not self.user_cloud.has_service('compute'):
            return None

        flavors = self.user_cloud.list_flavors(get_extra=False)

        flavor_name = os.environ.get('OPENSTACKSDK_FLAVOR')

        if not flavor_name:
            flavor_name = _get_resource_value('flavor_name')

        if flavor_name:
            for flavor in flavors:
                if flavor.name == flavor_name:
                    return flavor

            raise self.failureException(
                "Cloud does not have flavor '%s'",
                flavor_name,
            )

        # Enable running functional tests against RAX, which requires
        # performance flavors be used for boot from volume

        for flavor in sorted(flavors, key=operator.attrgetter('ram')):
            if 'performance' in flavor.name:
                return flavor

        # Otherwise, pick the smallest flavor with a ephemeral disk configured

        for flavor in sorted(flavors, key=operator.attrgetter('ram')):
            if flavor.disk:
                return flavor

        raise self.failureException('No sensible flavor found')

    def _pick_image(self):
        """Pick a sensible image to run tests with.

        This returns None if the image service is not present.
        """
        if not self.user_cloud.has_service('image'):
            return None

        images = self.user_cloud.list_images()

        image_name = os.environ.get('OPENSTACKSDK_IMAGE')

        if not image_name:
            image_name = _get_resource_value('image_name')

        if image_name:
            for image in images:
                if image.name == image_name:
                    return image

            raise self.failureException(
                "Cloud does not have image '%s'",
                image_name,
            )

        for image in images:
            if image.name.startswith('cirros') and image.name.endswith('-uec'):
                return image

        for image in images:
            if (
                image.name.startswith('cirros')
                and image.disk_format == 'qcow2'
            ):
                return image

        for image in images:
            if image.name.lower().startswith('ubuntu'):
                return image
        for image in images:
            if image.name.lower().startswith('centos'):
                return image

        raise self.failureException('No sensible image found')

    def addEmptyCleanup(self, func, *args, **kwargs):
        def cleanup():
            result = func(*args, **kwargs)
            self.assertIsNone(result)

        self.addCleanup(cleanup)

    def require_service(self, service_type, min_microversion=None, **kwargs):
        """Method to check whether a service exists

        Usage::

            class TestMeter(base.BaseFunctionalTest):
                def setUp(self):
                    super(TestMeter, self).setUp()
                    self.require_service('metering')

        :returns: True if the service exists, otherwise False.
        """
        if not self.operator_cloud.has_service(service_type):
            self.skipTest(f'Service {service_type} not found in cloud')

        if not min_microversion:
            return

        data = self.operator_cloud.session.get_endpoint_data(
            service_type=service_type, **kwargs
        )

        if not data or not (
            data.min_microversion
            and data.max_microversion
            and discover.version_between(
                data.min_microversion,
                data.max_microversion,
                min_microversion,
            )
        ):
            self.skipTest(
                f'Service {service_type} does not provide microversion '
                f'{min_microversion}'
            )

    def getUniqueString(self, prefix=None):
        """Generate unique resource name"""
        # Globally unique names can only rely on some form of uuid
        # unix_t is also used to easier determine orphans when running real
        # functional tests on a real cloud
        return (
            prefix if prefix else ''
        ) + f"{int(time.time())}-{uuid.uuid4().hex}"

    def create_temporary_project(self):
        """Create a new temporary project.

        This is useful for tests that modify things like quotas, which would
        cause issues for other tests.
        """
        project_name = self.getUniqueString('project-')
        project = self.operator_cloud.get_project(project_name)
        if not project:
            params = {
                'name': project_name,
                'description': f'Temporary project created for {self.id()}',
                # assume identity API v3 for now
                'domain_id': self.operator_cloud.get_domain('default')['id'],
            }
            project = self.operator_cloud.create_project(**params)

        # Grant the current user access to the project
        user_id = self.operator_cloud.current_user_id
        role_assignment = self.operator_cloud.list_role_assignments(
            {'user': user_id, 'project': project['id']}
        )
        if not role_assignment:
            self.operator_cloud.grant_role(
                'member', user=user_id, project=project['id'], wait=True
            )

        self.addCleanup(self._delete_temporary_project, project)

        return project

    def _delete_temporary_project(self, project):
        self.operator_cloud.revoke_role(
            'member',
            user=self.operator_cloud.current_user_id,
            project=project.id,
        )
        self.operator_cloud.delete_project(project.id)


class KeystoneBaseFunctionalTest(BaseFunctionalTest):
    def setUp(self):
        super().setUp()

        # we only support v3, since v2 was deprecated in Queens (2018)

        if not self.user_cloud.has_service('identity', '3'):
            self.skipTest('identity service not supported by cloud')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
from unittest import mock

from keystoneauth1 import adapter

from openstack.block_storage.v2 import snapshot
from openstack.tests.unit import base


FAKE_ID = "ffa9bc5e-1172-4021-acaf-cdcd78a9584d"
FAKE_VOLUME_ID = "5aa119a8-d25b-45a7-8d1b-88e127885635"

SNAPSHOT = {
    "status": "creating",
    "description": "Daily backup",
    "created_at": "2015-03-09T12:14:57.233772",
    "updated_at": None,
    "metadata": {},
    "volume_id": "5aa119a8-d25b-45a7-8d1b-88e127885635",
    "size": 1,
    "id": FAKE_ID,
    "name": "snap-001",
    "force": "true",
}

DETAILS = {
    "os-extended-snapshot-attributes:progress": "100%",
    "os-extended-snapshot-attributes:project_id": "0c2eba2c5af04d3f9e9d0d410b371fde",  # noqa: E501
}

DETAILED_SNAPSHOT = SNAPSHOT.copy()
DETAILED_SNAPSHOT.update(**DETAILS)


class TestSnapshot(base.TestCase):
    def test_basic(self):
        sot = snapshot.Snapshot(SNAPSHOT)
        self.assertEqual("snapshot", sot.resource_key)
        self.assertEqual("snapshots", sot.resources_key)
        self.assertEqual("/snapshots", sot.base_path)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

        self.assertDictEqual(
            {
                "name": "name",
                "status": "status",
                "all_projects": "all_tenants",
                "volume_id": "volume_id",
                "limit": "limit",
                "marker": "marker",
            },
            sot._query_mapping._mapping,
        )

    def test_create_basic(self):
        sot = snapshot.Snapshot(**SNAPSHOT)
        self.assertEqual(SNAPSHOT["id"], sot.id)
        self.assertEqual(SNAPSHOT["status"], sot.status)
        self.assertEqual(SNAPSHOT["created_at"], sot.created_at)
        self.assertEqual(SNAPSHOT["updated_at"], sot.updated_at)
        self.assertEqual(SNAPSHOT["metadata"], sot.metadata)
        self.assertEqual(SNAPSHOT["volume_id"], sot.volume_id)
        self.assertEqual(SNAPSHOT["size"], sot.size)
        self.assertEqual(SNAPSHOT["name"], sot.name)
        self.assertTrue(sot.is_forced)


class TestSnapshotActions(base.TestCase):
    def setUp(self):
        super().setUp()
        self.resp = mock.Mock()
        self.resp.body = None
        self.resp.json = mock.Mock(return_value=self.resp.body)
        self.resp.headers = {}
        self.resp.status_code = 202

        self.sess = mock.Mock(spec=adapter.Adapter)
        self.sess.get = mock.Mock()
        self.sess.post = mock.Mock(return_value=self.resp)

    def test_reset_status(self):
        sot = snapshot.Snapshot(**SNAPSHOT)

        self.assertIsNone(sot.reset_status(self.sess, 'new_status'))

        url = f'snapshots/{FAKE_ID}/action'
        body = {'os-reset_status': {'status': 'new_status'}}
        self.sess.post.assert_called_with(url, json=body)

    def test_manage(self):
        resp = mock.Mock()
        resp.body = {'snapshot': copy.deepcopy(SNAPSHOT)}
        resp.json = mock.Mock(return_value=resp.body)
        resp.headers = {}
        resp.status_code = 202

        self.sess.post = mock.Mock(return_value=resp)

        sot = snapshot.Snapshot.manage(
            self.sess, volume_id=FAKE_VOLUME_ID, ref=FAKE_ID
        )

        self.assertIsNotNone(sot)

        url = '/os-snapshot-manage'
        body = {
            'snapshot': {
                'volume_id': FAKE_VOLUME_ID,
                'ref': FAKE_ID,
                'name': None,
                'description': None,
                'metadata': None,
            }
        }
        self.sess.post.assert_called_with(url, json=body)

    def test_unmanage(self):
        sot = snapshot.Snapshot(**SNAPSHOT)

        self.assertIsNone(sot.unmanage(self.sess))

        url = f'snapshots/{FAKE_ID}/action'
        body = {'os-unmanage': None}
        self.sess.post.assert_called_with(url, json=body)



# Copyright(c) 2018 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from openstack import resource


class ProgressDetailsItem(resource.Resource):
    #: The timestamp of recovery workflow task.
    timestamp = resource.Body("timestamp")
    #: The message of recovery workflow task.
    message = resource.Body("message")
    #: The progress of recovery workflow task.
    progress = resource.Body("progress")


class RecoveryWorkflowDetailItem(resource.Resource):
    #: The progress of recovery workflow.
    progress = resource.Body("progress")
    #: The name of recovery workflow.
    name = resource.Body("name")
    #: The state of recovery workflow.
    state = resource.Body("state")
    #: The progress details of this recovery workflow.
    progress_details = resource.Body(
        "progress_details", type=list, list_type=ProgressDetailsItem
    )


class Notification(resource.Resource):
    resource_key = "notification"
    resources_key = "notifications"
    base_path = "/notifications"

    # capabilities
    # 1] GET /v1/notifications
    # 2] GET /v1/notifications/<notification_uuid>
    # 3] POST /v1/notifications
    allow_list = True
    allow_fetch = True
    allow_create = True
    allow_commit = False
    allow_delete = False

    #: A ID of representing this notification.
    id = resource.Body("id")
    #: A Uuid of representing this notification.
    notification_uuid = resource.Body("notification_uuid")
    #: A created time of representing this notification.
    created_at = resource.Body("created_at")
    #: A latest updated time of representing this notification.
    updated_at = resource.Body("updated_at")
    #: The type of failure. Valuse values include ''COMPUTE_HOST'',
    #: ''VM'', ''PROCESS''
    type = resource.Body("type")
    #: The hostname of this notification.
    hostname = resource.Body("hostname")
    #: The status for this notitication.
    status = resource.Body("status")
    #: The generated_time for this notitication.
    generated_time = resource.Body("generated_time")
    #: The payload of this notification.
    payload = resource.Body("payload")
    #: The source host uuid of this notification.
    source_host_uuid = resource.Body("source_host_uuid")
    #: The recovery workflow details of this notification.
    recovery_workflow_details = resource.Body(
        "recovery_workflow_details",
        type=list,
        list_type=RecoveryWorkflowDetailItem,
    )

    _query_mapping = resource.QueryParameters(
        "sort_key",
        "sort_dir",
        source_host_uuid="source_host_uuid",
        type="type",
        status="status",
        generated_since="generated-since",
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
# # Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.block_storage.v3 import resource_filter
from openstack.tests.unit import base

RESOURCE_FILTER = {
    'filters': [
        'name',
        'status',
        'image_metadata',
        'bootable',
        'migration_status',
    ],
    'resource': 'volume',
}


class TestResourceFilter(base.TestCase):
    def test_basic(self):
        resource = resource_filter.ResourceFilter()
        self.assertEqual('resource_filters', resource.resources_key)
        self.assertEqual('/resource_filters', resource.base_path)
        self.assertFalse(resource.allow_create)
        self.assertFalse(resource.allow_fetch)
        self.assertFalse(resource.allow_commit)
        self.assertFalse(resource.allow_delete)
        self.assertTrue(resource.allow_list)

        self.assertDictEqual(
            {
                "resource": "resource",
            },
            resource._query_mapping._mapping,
        )

    def test_make_resource_filter(self):
        resource = resource_filter.ResourceFilter(**RESOURCE_FILTER)
        self.assertEqual(RESOURCE_FILTER['filters'], resource.filters)
        self.assertEqual(RESOURCE_FILTER['resource'], resource.resource)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from unittest import mock

from keystoneauth1 import adapter

from openstack import exceptions
from openstack.image.v2 import metadef_namespace
from openstack.tests.unit import base


EXAMPLE = {
    'display_name': 'Cinder Volume Type',
    'created_at': '2022-08-24T17:46:24Z',
    'protected': True,
    'namespace': 'OS::Cinder::Volumetype',
    'description': (
        'The Cinder volume type configuration option. Volume type '
        'assignment provides a mechanism not only to provide scheduling to a '
        'specific storage back-end, but also can be used to specify specific '
        'information for a back-end storage device to act upon.'
    ),
    'visibility': 'public',
    'owner': 'admin',
    'resource_type_associations': [
        {
            'name': 'OS::Glance::Image',
            'prefix': 'cinder_',
            'created_at': '2022-08-24T17:46:24Z',
        },
    ],
}


class TestMetadefNamespace(base.TestCase):
    def test_basic(self):
        sot = metadef_namespace.MetadefNamespace()
        self.assertIsNone(sot.resource_key)
        self.assertEqual('namespaces', sot.resources_key)
        self.assertEqual('/metadefs/namespaces', sot.base_path)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_list)
        self.assertTrue(sot.allow_delete)

    def test_make_it(self):
        sot = metadef_namespace.MetadefNamespace(**EXAMPLE)
        self.assertEqual(EXAMPLE['namespace'], sot.namespace)
        self.assertEqual(EXAMPLE['visibility'], sot.visibility)
        self.assertEqual(EXAMPLE['owner'], sot.owner)
        self.assertEqual(EXAMPLE['created_at'], sot.created_at)
        self.assertEqual(EXAMPLE['protected'], sot.is_protected)
        self.assertEqual(EXAMPLE['display_name'], sot.display_name)
        self.assertEqual(
            EXAMPLE['resource_type_associations'],
            sot.resource_type_associations,
        )
        self.assertDictEqual(
            {
                'limit': 'limit',
                'marker': 'marker',
                'resource_types': 'resource_types',
                'sort_dir': 'sort_dir',
                'sort_key': 'sort_key',
                'visibility': 'visibility',
            },
            sot._query_mapping._mapping,
        )

    @mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
    def test_delete_all_properties(self):
        sot = metadef_namespace.MetadefNamespace(**EXAMPLE)
        session = mock.Mock(spec=adapter.Adapter)
        sot._translate_response = mock.Mock()
        sot.delete_all_properties(session)
        session.delete.assert_called_with(
            'metadefs/namespaces/OS::Cinder::Volumetype/properties'
        )

    @mock.patch.object(exceptions, 'raise_from_response', mock.Mock())
    def test_delete_all_objects(self):
        sot = metadef_namespace.MetadefNamespace(**EXAMPLE)
        session = mock.Mock(spec=adapter.Adapter)
        sot._translate_response = mock.Mock()
        sot.delete_all_objects(session)
        session.delete.assert_called_with(
            'metadefs/namespaces/OS::Cinder::Volumetype/objects'
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class TapFlow(resource.Resource):
    """Tap Flow"""

    resource_key = 'tap_flow'
    resources_key = 'tap_flows'
    base_path = '/taas/tap_flows'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    _allow_unknown_attrs_in_body = True

    _query_mapping = resource.QueryParameters(
        "sort_key",
        "sort_dir",
        'name',
        'project_id',
    )

    # Properties
    #: The ID of the tap flow.
    id = resource.Body('id')
    #: The tap flow's name.
    name = resource.Body('name')
    #: The tap flow's description.
    description = resource.Body('description')
    #: The ID of the project that owns the tap flow.
    project_id = resource.Body('project_id', alias='tenant_id')
    #: Tenant_id (deprecated attribute).
    tenant_id = resource.Body('tenant_id', deprecated=True)
    #: The id of the tap_service with which the tap flow is associated
    tap_service_id = resource.Body('tap_service_id')
    #: The direction of the tap flow.
    direction = resource.Body('direction')
    #: The status for the tap flow.
    status = resource.Body('status')
    #: The id of the port the tap flow is associated with
    source_port = resource.Body('source_port')



# Copyright 2010-2011 OpenStack Foundation
# Copyright (c) 2013 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import os
import tempfile
import time
import urllib
import uuid

import fixtures
from keystoneauth1 import loading as ks_loading
from oslo_config import cfg
from requests import structures
from requests_mock.contrib import fixture as rm_fixture

import openstack.cloud
import openstack.config as occ
import openstack.connection
from openstack.fixture import connection as os_fixture
from openstack.tests import base
from openstack.tests import fakes


_ProjectData = collections.namedtuple(
    'ProjectData',
    'project_id, project_name, enabled, domain_id, description, '
    'parent_id, json_response, json_request',
)


_UserData = collections.namedtuple(
    'UserData',
    'user_id, password, name, email, description, domain_id, enabled, '
    'json_response, json_request',
)


_GroupData = collections.namedtuple(
    'GroupData',
    'group_id, group_name, domain_id, description, json_response, '
    'json_request',
)


_DomainData = collections.namedtuple(
    'DomainData',
    'domain_id, domain_name, description, json_response, json_request',
)


_ServiceData = collections.namedtuple(
    'Servicedata',
    'service_id, service_name, service_type, description, enabled, '
    'json_response_v3, json_response_v2, json_request',
)


_EndpointDataV3 = collections.namedtuple(
    'EndpointData',
    'endpoint_id, service_id, interface, region_id, url, enabled, '
    'json_response, json_request',
)


# NOTE(notmorgan): Shade does not support domain-specific roles
# This should eventually be fixed if it becomes a main-stream feature.
_RoleData = collections.namedtuple(
    'RoleData', 'role_id, role_name, json_response, json_request'
)


class TestCase(base.TestCase):
    strict_cloud = False

    def setUp(self, cloud_config_fixture='clouds.yaml'):
        """Run before each test method to initialize test environment."""

        super().setUp()

        # Sleeps are for real testing, but unit tests shouldn't need them
        realsleep = time.sleep

        def _nosleep(seconds):
            return realsleep(seconds * 0.0001)

        self.sleep_fixture = self.useFixture(
            fixtures.MonkeyPatch('time.sleep', _nosleep)
        )
        self.fixtures_directory = 'openstack/tests/unit/fixtures'
        self.os_fixture = self.useFixture(
            os_fixture.ConnectionFixture(project_id=fakes.PROJECT_ID)
        )

        # Isolate openstack.config from test environment
        self.os_cloud_fixture = self.useFixture(
            fixtures.EnvironmentVariable('OS_CLOUD'),
        )
        config = tempfile.NamedTemporaryFile(delete=False)
        cloud_path = os.path.join(
            self.fixtures_directory,
            'clouds',
            cloud_config_fixture,
        )
        with open(cloud_path, 'rb') as f:
            content = f.read()
            config.write(content)
        config.close()

        vendor = tempfile.NamedTemporaryFile(delete=False)
        vendor.write(b'{}')
        vendor.close()

        self.config = occ.OpenStackConfig(
            config_files=[config.name],
            vendor_files=[vendor.name],
            secure_files=['non-existant'],
        )

        self.oslo_config_dict = {
            # All defaults for nova
            'nova': {},
            # monasca-api not in the service catalog
            'monasca-api': {},
            # Overrides for heat
            'heat': {
                'region_name': 'SpecialRegion',
                'interface': 'internal',
                'endpoint_override': 'https://example.org:8888/heat/v2',
            },
            # test a service with dashes
            'ironic_inspector': {
                'endpoint_override': 'https://example.org:5050',
            },
        }

        # FIXME(notmorgan): Convert the uri_registry, discovery.json, and
        # use of keystone_v3/v2 to a proper fixtures.Fixture. For now this
        # is acceptable, but eventually this should become it's own fixture
        # that encapsulates the registry, registering the URIs, and
        # assert_calls (and calling assert_calls every test case that uses
        # it on cleanup). Subclassing here could be 100% eliminated in the
        # future allowing any class to simply
        # self.useFixture(openstack.cloud.RequestsMockFixture) and get all
        # the benefits.

        # NOTE(notmorgan): use an ordered dict here to ensure we preserve the
        # order in which items are added to the uri_registry. This makes
        # the behavior more consistent when dealing with ensuring the
        # requests_mock uri/query_string matchers are ordered and parse the
        # request in the correct orders.
        self._uri_registry = collections.OrderedDict()
        self.discovery_json = os.path.join(
            self.fixtures_directory, 'discovery.json'
        )
        self.use_keystone_v3()
        self.__register_uris_called = False

    def _load_ks_cfg_opts(self):
        conf = cfg.ConfigOpts()
        for group, opts in self.oslo_config_dict.items():
            conf.register_group(cfg.OptGroup(group))
            if opts is not None:
                ks_loading.register_adapter_conf_options(conf, group)
                for name, val in opts.items():
                    conf.set_override(name, val, group=group)
        return conf

    # TODO(shade) Update this to handle service type aliases
    def get_mock_url(
        self,
        service_type,
        interface='public',
        resource=None,
        append=None,
        base_url_append=None,
        qs_elements=None,
    ):
        endpoint_url = self.cloud.endpoint_for(
            service_type=service_type, interface=interface
        )
        # Strip trailing slashes, so as not to produce double-slashes below
        if endpoint_url.endswith('/'):
            endpoint_url = endpoint_url[:-1]
        to_join = [endpoint_url]
        qs = ''
        if base_url_append:
            to_join.append(base_url_append)
        if resource:
            to_join.append(resource)
        if append:
            to_join.extend([urllib.parse.quote(i) for i in append])
        if qs_elements is not None:
            qs = '?{}'.format('&'.join(qs_elements))
        return '{uri}{qs}'.format(uri='/'.join(to_join), qs=qs)

    def mock_for_keystone_projects(
        self,
        project=None,
        v3=True,
        list_get=False,
        id_get=False,
        project_list=None,
        project_count=None,
    ):
        if project:
            assert not (project_list or project_count)
        elif project_list:
            assert not (project or project_count)
        elif project_count:
            assert not (project or project_list)
        else:
            raise Exception(
                'Must specify a project, project_list, or project_count'
            )
        assert list_get or id_get

        base_url_append = 'v3' if v3 else None
        if project:
            project_list = [project]
        elif project_count:
            # Generate multiple projects
            project_list = [
                self._get_project_data(v3=v3) for c in range(0, project_count)
            ]
        uri_mock_list = []
        if list_get:
            uri_mock_list.append(
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        service_type='identity',
                        interface='admin',
                        resource='projects',
                        base_url_append=base_url_append,
                    ),
                    status_code=200,
                    json={
                        'projects': [
                            p.json_response['project'] for p in project_list
                        ]
                    },
                )
            )
        if id_get:
            for p in project_list:
                uri_mock_list.append(
                    dict(
                        method='GET',
                        uri=self.get_mock_url(
                            service_type='identity',
                            interface='admin',
                            resource='projects',
                            append=[p.project_id],
                            base_url_append=base_url_append,
                        ),
                        status_code=200,
                        json=p.json_response,
                    )
                )
        self.__do_register_uris(uri_mock_list)
        return project_list

    def _get_project_data(
        self,
        project_name=None,
        enabled=None,
        domain_id=None,
        description=None,
        v3=True,
        project_id=None,
        parent_id=None,
    ):
        project_name = project_name or self.getUniqueString('projectName')
        project_id = uuid.UUID(project_id or uuid.uuid4().hex).hex
        if parent_id:
            parent_id = uuid.UUID(parent_id).hex
        response = {'id': project_id, 'name': project_name}
        request = {'name': project_name}
        domain_id = (domain_id or uuid.uuid4().hex) if v3 else None
        if domain_id:
            request['domain_id'] = domain_id
            response['domain_id'] = domain_id
        if enabled is not None:
            enabled = bool(enabled)
            response['enabled'] = enabled
            request['enabled'] = enabled
        if parent_id:
            request['parent_id'] = parent_id
            response['parent_id'] = parent_id
        response.setdefault('enabled', True)
        request.setdefault('enabled', True)
        if description:
            response['description'] = description
            request['description'] = description
        request.setdefault('description', None)
        return _ProjectData(
            project_id,
            project_name,
            enabled,
            domain_id,
            description,
            parent_id,
            {'project': response},
            {'project': request},
        )

    def _get_group_data(self, name=None, domain_id=None, description=None):
        group_id = uuid.uuid4().hex
        name = name or self.getUniqueString('groupname')
        domain_id = uuid.UUID(domain_id or uuid.uuid4().hex).hex
        response = {'id': group_id, 'name': name, 'domain_id': domain_id}
        request = {'name': name, 'domain_id': domain_id}
        if description is not None:
            response['description'] = description
            request['description'] = description

        return _GroupData(
            group_id,
            name,
            domain_id,
            description,
            {'group': response},
            {'group': request},
        )

    def _get_user_data(self, name=None, password=None, **kwargs):
        name = name or self.getUniqueString('username')
        password = password or self.getUniqueString('user_password')
        user_id = uuid.uuid4().hex

        response = {'name': name, 'id': user_id}
        request = {'name': name, 'password': password}

        if kwargs.get('domain_id'):
            kwargs['domain_id'] = uuid.UUID(kwargs['domain_id']).hex
            response['domain_id'] = kwargs.pop('domain_id')
            request['domain_id'] = response['domain_id']

        response['email'] = kwargs.pop('email', None)
        request['email'] = response['email']

        response['enabled'] = kwargs.pop('enabled', True)
        request['enabled'] = response['enabled']

        response['description'] = kwargs.pop('description', None)
        if response['description']:
            request['description'] = response['description']

        self.assertIs(
            0,
            len(kwargs),
            message='extra key-word args received on _get_user_data',
        )

        return _UserData(
            user_id,
            password,
            name,
            response['email'],
            response['description'],
            response.get('domain_id'),
            response.get('enabled'),
            {'user': response},
            {'user': request},
        )

    def _get_domain_data(
        self, domain_name=None, description=None, enabled=None
    ):
        domain_id = uuid.uuid4().hex
        domain_name = domain_name or self.getUniqueString('domainName')
        response = {'id': domain_id, 'name': domain_name}
        request = {'name': domain_name}
        if enabled is not None:
            request['enabled'] = bool(enabled)
            response['enabled'] = bool(enabled)
        if description:
            response['description'] = description
            request['description'] = description
        response.setdefault('enabled', True)
        return _DomainData(
            domain_id,
            domain_name,
            description,
            {'domain': response},
            {'domain': request},
        )

    def _get_service_data(
        self, type=None, name=None, description=None, enabled=True
    ):
        service_id = uuid.uuid4().hex
        name = name or f'name-{uuid.uuid4().hex}'
        type = type or uuid.uuid4().hex

        response = {
            'id': service_id,
            'name': name,
            'type': type,
            'enabled': enabled,
        }
        if description is not None:
            response['description'] = description
        request = response.copy()
        request.pop('id')
        return _ServiceData(
            service_id,
            name,
            type,
            description,
            enabled,
            {'service': response},
            {'OS-KSADM:service': response},
            request,
        )

    def _get_endpoint_v3_data(
        self,
        service_id=None,
        region=None,
        url=None,
        interface=None,
        enabled=True,
    ):
        endpoint_id = uuid.uuid4().hex
        service_id = service_id or uuid.uuid4().hex
        region = region or uuid.uuid4().hex
        url = url or 'https://example.com/'
        interface = interface or uuid.uuid4().hex

        response = {
            'id': endpoint_id,
            'service_id': service_id,
            'region_id': region,
            'interface': interface,
            'url': url,
            'enabled': enabled,
        }
        request = response.copy()
        request.pop('id')
        return _EndpointDataV3(
            endpoint_id,
            service_id,
            interface,
            region,
            url,
            enabled,
            {'endpoint': response},
            {'endpoint': request},
        )

    def _get_role_data(self, role_name=None):
        role_id = uuid.uuid4().hex
        role_name = role_name or uuid.uuid4().hex
        request = {'name': role_name}
        response = request.copy()
        response['id'] = role_id
        return _RoleData(
            role_id, role_name, {'role': response}, {'role': request}
        )

    def use_broken_keystone(self):
        self.adapter = self.useFixture(rm_fixture.Fixture())
        self.calls = []
        self._uri_registry.clear()
        self.__do_register_uris(
            [
                dict(
                    method='GET',
                    uri='https://identity.example.com/',
                    text=open(self.discovery_json).read(),
                ),
                dict(
                    method='POST',
                    uri='https://identity.example.com/v3/auth/tokens',
                    status_code=400,
                ),
            ]
        )
        self._make_test_cloud(identity_api_version='3')

    def use_nothing(self):
        self.calls = []
        self._uri_registry.clear()

    def get_keystone_v3_token(
        self,
        project_name='admin',
    ):
        return dict(
            method='POST',
            uri='https://identity.example.com/v3/auth/tokens',
            headers={'X-Subject-Token': self.getUniqueString('KeystoneToken')},
            json=self.os_fixture.v3_token,
            validate=dict(
                json={
                    'auth': {
                        'identity': {
                            'methods': ['password'],
                            'password': {
                                'user': {
                                    'domain': {
                                        'name': 'default',
                                    },
                                    'name': 'admin',
                                    'password': 'password',
                                }
                            },
                        },
                        'scope': {
                            'project': {
                                'domain': {'name': 'default'},
                                'name': project_name,
                            }
                        },
                    }
                }
            ),
        )

    def get_keystone_discovery(self):
        with open(self.discovery_json) as discovery_file:
            return dict(
                method='GET',
                uri='https://identity.example.com/',
                text=discovery_file.read(),
            )

    def use_keystone_v3(self):
        self.adapter = self.useFixture(rm_fixture.Fixture())
        self.calls = []
        self._uri_registry.clear()
        self.__do_register_uris(
            [
                self.get_keystone_discovery(),
                self.get_keystone_v3_token(),
            ]
        )
        self._make_test_cloud(identity_api_version='3')

    def use_keystone_v2(self):
        self.adapter = self.useFixture(rm_fixture.Fixture())
        self.calls = []
        self._uri_registry.clear()

        self.__do_register_uris(
            [
                self.get_keystone_discovery(),
                dict(
                    method='POST',
                    uri='https://identity.example.com/v2.0/tokens',
                    json=self.os_fixture.v2_token,
                ),
            ]
        )

        self._make_test_cloud(
            cloud_name='_test_cloud_v2_', identity_api_version='2.0'
        )

    def _make_test_cloud(self, cloud_name='_test_cloud_', **kwargs):
        test_cloud = os.environ.get('OPENSTACKSDK_OS_CLOUD', cloud_name)
        self.cloud_config = self.config.get_one(
            cloud=test_cloud, validate=True, **kwargs
        )
        self.cloud = openstack.connection.Connection(
            config=self.cloud_config, strict=self.strict_cloud
        )

    def get_cinder_discovery_mock_dict(
        self,
        block_storage_version_json='block-storage-version.json',
        block_storage_discovery_url='https://block-storage.example.com/',
    ):
        discovery_fixture = os.path.join(
            self.fixtures_directory, block_storage_version_json
        )
        return dict(
            method='GET',
            uri=block_storage_discovery_url,
            text=open(discovery_fixture).read(),
        )

    def get_glance_discovery_mock_dict(
        self,
        image_version_json='image-version.json',
        image_discovery_url='https://image.example.com/',
    ):
        discovery_fixture = os.path.join(
            self.fixtures_directory, image_version_json
        )
        return dict(
            method='GET',
            uri=image_discovery_url,
            status_code=300,
            text=open(discovery_fixture).read(),
        )

    def get_nova_discovery_mock_dict(
        self,
        compute_version_json='compute-version.json',
        compute_discovery_url='https://compute.example.com/v2.1/',
    ):
        discovery_fixture = os.path.join(
            self.fixtures_directory, compute_version_json
        )
        return dict(
            method='GET',
            uri=compute_discovery_url,
            text=open(discovery_fixture).read(),
        )

    def get_placement_discovery_mock_dict(
        self, discovery_fixture='placement.json'
    ):
        discovery_fixture = os.path.join(
            self.fixtures_directory, discovery_fixture
        )
        return dict(
            method='GET',
            uri="https://placement.example.com/",
            text=open(discovery_fixture).read(),
        )

    def get_designate_discovery_mock_dict(self):
        discovery_fixture = os.path.join(self.fixtures_directory, "dns.json")
        return dict(
            method='GET',
            uri="https://dns.example.com/",
            text=open(discovery_fixture).read(),
        )

    def get_ironic_discovery_mock_dict(self):
        discovery_fixture = os.path.join(
            self.fixtures_directory, "baremetal.json"
        )
        return dict(
            method='GET',
            uri="https://baremetal.example.com/",
            text=open(discovery_fixture).read(),
        )

    def get_senlin_discovery_mock_dict(self):
        discovery_fixture = os.path.join(
            self.fixtures_directory, "clustering.json"
        )
        return dict(
            method='GET',
            uri="https://clustering.example.com/",
            text=open(discovery_fixture).read(),
        )

    def use_compute_discovery(
        self,
        compute_version_json='compute-version.json',
        compute_discovery_url='https://compute.example.com/v2.1/',
    ):
        self.__do_register_uris(
            [
                self.get_nova_discovery_mock_dict(
                    compute_version_json, compute_discovery_url
                ),
            ]
        )

    def get_cyborg_discovery_mock_dict(self):
        discovery_fixture = os.path.join(
            self.fixtures_directory, "accelerator.json"
        )
        return dict(
            method='GET',
            uri="https://accelerator.example.com/",
            text=open(discovery_fixture).read(),
        )

    def get_manila_discovery_mock_dict(self):
        discovery_fixture = os.path.join(
            self.fixtures_directory, "shared-file-system.json"
        )
        return dict(
            method='GET',
            uri="https://shared-file-system.example.com/",
            text=open(discovery_fixture).read(),
        )

    def use_glance(
        self,
        image_version_json='image-version.json',
        image_discovery_url='https://image.example.com/',
    ):
        # NOTE(notmorgan): This method is only meant to be used in "setUp"
        # where the ordering of the url being registered is tightly controlled
        # if the functionality of .use_glance is meant to be used during an
        # actual test case, use .get_glance_discovery_mock and apply to the
        # right location in the mock_uris when calling .register_uris
        self.__do_register_uris(
            [
                self.get_glance_discovery_mock_dict(
                    image_version_json, image_discovery_url
                )
            ]
        )

    def use_cinder(self):
        self.__do_register_uris([self.get_cinder_discovery_mock_dict()])

    def use_placement(self, **kwargs):
        self.__do_register_uris(
            [self.get_placement_discovery_mock_dict(**kwargs)]
        )

    def use_designate(self):
        # NOTE(slaweq): This method is only meant to be used in "setUp"
        # where the ordering of the url being registered is tightly controlled
        # if the functionality of .use_designate is meant to be used during an
        # actual test case, use .get_designate_discovery_mock and apply to the
        # right location in the mock_uris when calling .register_uris
        self.__do_register_uris([self.get_designate_discovery_mock_dict()])

    def use_ironic(self):
        # NOTE(TheJulia): This method is only meant to be used in "setUp"
        # where the ordering of the url being registered is tightly controlled
        # if the functionality of .use_ironic is meant to be used during an
        # actual test case, use .get_ironic_discovery_mock and apply to the
        # right location in the mock_uris when calling .register_uris
        self.__do_register_uris([self.get_ironic_discovery_mock_dict()])

    def use_senlin(self):
        # NOTE(elachance): This method is only meant to be used in "setUp"
        # where the ordering of the url being registered is tightly controlled
        # if the functionality of .use_senlin is meant to be used during an
        # actual test case, use .get_senlin_discovery_mock and apply to the
        # right location in the mock_uris when calling .register_uris
        self.__do_register_uris([self.get_senlin_discovery_mock_dict()])

    def use_cyborg(self):
        # NOTE(s_shogo): This method is only meant to be used in "setUp"
        # where the ordering of the url being registered is tightly controlled
        # if the functionality of .use_cyborg is meant to be used during an
        # actual test case, use .get_cyborg_discovery_mock and apply to the
        # right location in the mock_uris when calling .register_uris
        self.__do_register_uris([self.get_cyborg_discovery_mock_dict()])

    def use_manila(self):
        # NOTE(gouthamr): This method is only meant to be used in "setUp"
        # where the ordering of the url being registered is tightly controlled
        # if the functionality of .use_manila is meant to be used during an
        # actual test case, use .get_manila_discovery_mock and apply to the
        # right location in the mock_uris when calling .register_uris
        self.__do_register_uris([self.get_manila_discovery_mock_dict()])

    def register_uris(self, uri_mock_list=None):
        """Mock a list of URIs and responses via requests mock.

        This method may be called only once per test-case to avoid odd
        and difficult to debug interactions. Discovery and Auth request mocking
        happens separately from this method.

        :param uri_mock_list: List of dictionaries that template out what is
                              passed to requests_mock fixture's `register_uri`.
                              Format is:
                                  {'method': <HTTP_METHOD>,
                                   'uri': <URI to be mocked>,
                                   ...
                                  }

                              Common keys to pass in the dictionary:
                                  * json: the json response (dict)
                                  * status_code: the HTTP status (int)
                                  * validate: The request body (dict) to
                                              validate with assert_calls
                              all key-word arguments that are valid to send to
                              requests_mock are supported.

                              This list should be in the order in which calls
                              are made. When `assert_calls` is executed, order
                              here will be validated. Duplicate URIs and
                              Methods are allowed and will be collapsed into a
                              single matcher. Each response will be returned
                              in order as the URI+Method is hit.
        :type uri_mock_list: list
        :return: None
        """
        assert not self.__register_uris_called
        self.__do_register_uris(uri_mock_list or [])
        self.__register_uris_called = True

    def __do_register_uris(self, uri_mock_list=None):
        for to_mock in uri_mock_list:
            kw_params = {
                k: to_mock.pop(k)
                for k in ('request_headers', 'complete_qs', '_real_http')
                if k in to_mock
            }

            method = to_mock.pop('method')
            uri = to_mock.pop('uri')
            # NOTE(notmorgan): make sure the delimiter is non-url-safe, in this
            # case "|" is used so that the split can be a bit easier on
            # maintainers of this code.
            key = f'{method}|{uri}|{kw_params}'
            validate = to_mock.pop('validate', {})
            valid_keys = {'json', 'headers', 'params', 'data'}
            invalid_keys = set(validate.keys()) - valid_keys
            if invalid_keys:
                raise TypeError(
                    f"Invalid values passed to validate: {invalid_keys}"
                )
            headers = structures.CaseInsensitiveDict(
                to_mock.pop('headers', {})
            )
            if 'content-type' not in headers:
                headers['content-type'] = 'application/json'

            if 'exc' not in to_mock:
                to_mock['headers'] = headers

            self.calls += [dict(method=method, url=uri, **validate)]
            self._uri_registry.setdefault(
                key, {'response_list': [], 'kw_params': kw_params}
            )
            if self._uri_registry[key]['kw_params'] != kw_params:
                raise AssertionError(
                    'PROGRAMMING ERROR: key-word-params '
                    'should be part of the uri_key and cannot change, '
                    'it will affect the matcher in requests_mock. '
                    '{old!r} != {new!r}'.format(
                        old=self._uri_registry[key]['kw_params'],
                        new=kw_params,
                    )
                )
            self._uri_registry[key]['response_list'].append(to_mock)

        for mocked, params in self._uri_registry.items():
            mock_method, mock_uri, _ignored = mocked.split('|', 2)
            self.adapter.register_uri(
                mock_method,
                mock_uri,
                params['response_list'],
                **params['kw_params'],
            )

    def assert_no_calls(self):
        # TODO(mordred) For now, creating the adapter for connections is
        # triggering catalog lookups. Make sure no_calls is only 2.
        # When we can make that on-demand through a descriptor object,
        # drop this to 0.
        self.assertEqual(2, len(self.adapter.request_history))

    def assert_calls(self, stop_after=None, do_count=True):
        for x, (call, history) in enumerate(
            zip(self.calls, self.adapter.request_history)
        ):
            if stop_after and x > stop_after:
                break

            call_uri_parts = urllib.parse.urlparse(call['url'])
            history_uri_parts = urllib.parse.urlparse(history.url)
            self.assertEqual(
                (
                    call['method'],
                    call_uri_parts.scheme,
                    call_uri_parts.netloc,
                    call_uri_parts.path,
                    call_uri_parts.params,
                    urllib.parse.parse_qs(call_uri_parts.query),
                ),
                (
                    history.method,
                    history_uri_parts.scheme,
                    history_uri_parts.netloc,
                    history_uri_parts.path,
                    history_uri_parts.params,
                    urllib.parse.parse_qs(history_uri_parts.query),
                ),
                (
                    f'REST mismatch on call {x}. '
                    f'Expected {call["method"]} {call["url"]}. '
                    f'Got {history.method} {history.url}. '
                    'NOTE: query string order differences wont cause mismatch'
                ),
            )
            if 'json' in call:
                self.assertEqual(
                    call['json'],
                    history.json(),
                    f'json content mismatch in call {x}',
                )
            # headers in a call isn't exhaustive - it's checking to make sure
            # a specific header or headers are there, not that they are the
            # only headers
            if 'headers' in call:
                for key, value in call['headers'].items():
                    self.assertEqual(
                        value,
                        history.headers[key],
                        f'header mismatch in call {x}',
                    )
        if do_count:
            self.assertEqual(
                len(self.calls),
                len(self.adapter.request_history),
                "Expected:\n{}'\nGot:\n{}".format(
                    '\n'.join(
                        [f'{c["method"]} {c["url"]}' for c in self.calls]
                    ),
                    '\n'.join(
                        [
                            f'{h.method} {h.url}'
                            for h in self.adapter.request_history
                        ]
                    ),
                ),
            )

    def assertResourceEqual(self, actual, expected, resource_type):
        """Helper for the assertEqual which compares Resource object against
        dictionary representing expected state.

        :param Resource actual: actual object.
        :param dict expected: dictionary representing expected object.
        :param class resource_type: class type to be applied for the expected
            resource.
        """
        return self.assertEqual(
            resource_type(**expected).to_dict(computed=False),
            actual.to_dict(computed=False),
        )

    def assertResourceListEqual(self, actual, expected, resource_type):
        """Helper for the assertEqual which compares Resource lists object
        against dictionary representing expected state.

        :param list actual: List of actual objects.
        :param listexpected: List of dictionaries representing expected
            objects.
        :param class resource_type: class type to be applied for the expected
            resource.
        """
        self.assertEqual(
            [resource_type(**f).to_dict(computed=False) for f in expected],
            [f.to_dict(computed=False) for f in actual],
        )


class IronicTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.use_ironic()
        self.uuid = str(uuid.uuid4())
        self.name = self.getUniqueString('name')

    def get_mock_url(self, **kwargs):
        kwargs.setdefault('service_type', 'baremetal')
        kwargs.setdefault('interface', 'public')
        kwargs.setdefault('base_url_append', 'v1')
        return super().get_mock_url(**kwargs)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
test_network
----------------------------------

Functional tests for network methods.
"""

from openstack import exceptions
from openstack.tests.functional import base


class TestNetwork(base.BaseFunctionalTest):
    def setUp(self):
        super().setUp()
        if not self.operator_cloud:
            self.skipTest("Operator cloud is required for this test")

        if not self.operator_cloud.has_service('network'):
            self.skipTest('Network service not supported by cloud')
        self.network_name = self.getUniqueString('network')
        self.addCleanup(self._cleanup_networks)

    def _cleanup_networks(self):
        exception_list = list()
        for network in self.operator_cloud.list_networks():
            if network['name'].startswith(self.network_name):
                try:
                    self.operator_cloud.delete_network(network['name'])
                except Exception as e:
                    exception_list.append(str(e))
                    continue

        if exception_list:
            raise exceptions.SDKException('\n'.join(exception_list))

    def test_create_network_basic(self):
        net1 = self.operator_cloud.create_network(name=self.network_name)
        self.assertIn('id', net1)
        self.assertEqual(self.network_name, net1['name'])
        self.assertFalse(net1['shared'])
        self.assertFalse(net1['router:external'])
        self.assertTrue(net1['admin_state_up'])
        self.assertTrue(net1['port_security_enabled'])

    def test_get_network_by_id(self):
        net1 = self.operator_cloud.create_network(name=self.network_name)
        self.assertIn('id', net1)
        self.assertEqual(self.network_name, net1['name'])
        self.assertFalse(net1['shared'])
        self.assertFalse(net1['router:external'])
        self.assertTrue(net1['admin_state_up'])

        ret_net1 = self.operator_cloud.get_network_by_id(net1.id)
        self.assertIn('id', ret_net1)
        self.assertEqual(self.network_name, ret_net1['name'])
        self.assertFalse(ret_net1['shared'])
        self.assertFalse(ret_net1['router:external'])
        self.assertTrue(ret_net1['admin_state_up'])

    def test_create_network_advanced(self):
        net1 = self.operator_cloud.create_network(
            name=self.network_name,
            shared=True,
            external=True,
            admin_state_up=False,
        )
        self.assertIn('id', net1)
        self.assertEqual(self.network_name, net1['name'])
        self.assertTrue(net1['router:external'])
        self.assertTrue(net1['shared'])
        self.assertFalse(net1['admin_state_up'])

    def test_create_network_provider_flat(self):
        existing_public = self.operator_cloud.search_networks(
            filters={'provider:network_type': 'flat'}
        )
        if existing_public:
            self.skipTest('Physical network already allocated')
        net1 = self.operator_cloud.create_network(
            name=self.network_name,
            shared=True,
            provider={
                'physical_network': 'public',
                'network_type': 'flat',
            },
        )
        self.assertIn('id', net1)
        self.assertEqual(self.network_name, net1['name'])
        self.assertEqual('flat', net1['provider:network_type'])
        self.assertEqual('public', net1['provider:physical_network'])
        self.assertIsNone(net1['provider:segmentation_id'])

    def test_create_network_port_security_disabled(self):
        net1 = self.operator_cloud.create_network(
            name=self.network_name,
            port_security_enabled=False,
        )
        self.assertIn('id', net1)
        self.assertEqual(self.network_name, net1['name'])
        self.assertTrue(net1['admin_state_up'])
        self.assertFalse(net1['shared'])
        self.assertFalse(net1['router:external'])
        self.assertFalse(net1['port_security_enabled'])

    def test_list_networks_filtered(self):
        net1 = self.operator_cloud.create_network(name=self.network_name)
        self.assertIsNotNone(net1)
        net2 = self.operator_cloud.create_network(
            name=self.network_name + 'other'
        )
        self.assertIsNotNone(net2)
        match = self.operator_cloud.list_networks(
            filters=dict(name=self.network_name)
        )
        self.assertEqual(1, len(match))
        self.assertEqual(net1['name'], match[0]['name'])

    def test_update_network(self):
        net = self.operator_cloud.create_network(name=self.network_name)
        self.assertEqual(net.name, self.network_name)
        new_name = self.getUniqueString('network')
        net = self.operator_cloud.update_network(net.id, name=new_name)
        self.addCleanup(self.operator_cloud.delete_network, new_name)
        self.assertNotEqual(net.name, self.network_name)
        self.assertEqual(net.name, new_name)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from unittest import mock

import testtools

from openstack import exceptions
from openstack.network.v2 import network as _network
from openstack.tests.unit import base


class TestNeutronExtensions(base.TestCase):
    def test__neutron_extensions(self):
        body = [
            {
                "updated": "2014-06-1T10:00:00-00:00",
                "name": "Distributed Virtual Router",
                "links": [],
                "alias": "dvr",
                "description": "Enables configuration of Distributed Virtual Routers.",  # noqa: E501
            },
            {
                "updated": "2013-07-23T10:00:00-00:00",
                "name": "Allowed Address Pairs",
                "links": [],
                "alias": "allowed-address-pairs",
                "description": "Provides allowed address pairs",
            },
        ]
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'extensions']
                    ),
                    json=dict(extensions=body),
                )
            ]
        )
        extensions = self.cloud._neutron_extensions()
        self.assertEqual({'dvr', 'allowed-address-pairs'}, extensions)

        self.assert_calls()

    def test__neutron_extensions_fails(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'extensions']
                    ),
                    status_code=404,
                )
            ]
        )
        with testtools.ExpectedException(exceptions.NotFoundException):
            self.cloud._neutron_extensions()

        self.assert_calls()

    def test__has_neutron_extension(self):
        body = [
            {
                "updated": "2014-06-1T10:00:00-00:00",
                "name": "Distributed Virtual Router",
                "links": [],
                "alias": "dvr",
                "description": "Enables configuration of Distributed Virtual Routers.",  # noqa: E501
            },
            {
                "updated": "2013-07-23T10:00:00-00:00",
                "name": "Allowed Address Pairs",
                "links": [],
                "alias": "allowed-address-pairs",
                "description": "Provides allowed address pairs",
            },
        ]
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'extensions']
                    ),
                    json=dict(extensions=body),
                )
            ]
        )
        self.assertTrue(self.cloud._has_neutron_extension('dvr'))
        self.assert_calls()

    def test__has_neutron_extension_missing(self):
        body = [
            {
                "updated": "2014-06-1T10:00:00-00:00",
                "name": "Distributed Virtual Router",
                "links": [],
                "alias": "dvr",
                "description": "Enables configuration of Distributed Virtual Routers.",  # noqa: E501
            },
            {
                "updated": "2013-07-23T10:00:00-00:00",
                "name": "Allowed Address Pairs",
                "links": [],
                "alias": "allowed-address-pairs",
                "description": "Provides allowed address pairs",
            },
        ]
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'extensions']
                    ),
                    json=dict(extensions=body),
                )
            ]
        )
        self.assertFalse(self.cloud._has_neutron_extension('invalid'))
        self.assert_calls()


class TestNetworks(base.TestCase):
    mock_new_network_rep = {
        'provider:physical_network': None,
        'ipv6_address_scope': None,
        'revision_number': 3,
        'port_security_enabled': True,
        'provider:network_type': 'local',
        'id': '881d1bb7-a663-44c0-8f9f-ee2765b74486',
        'router:external': False,
        'availability_zone_hints': [],
        'availability_zones': [],
        'provider:segmentation_id': None,
        'ipv4_address_scope': None,
        'shared': False,
        'project_id': '861808a93da0484ea1767967c4df8a23',
        'status': 'ACTIVE',
        'subnets': [],
        'description': '',
        'tags': [],
        'updated_at': '2017-04-22T19:22:53Z',
        'is_default': False,
        'qos_policy_id': None,
        'name': 'netname',
        'admin_state_up': True,
        'created_at': '2017-04-22T19:22:53Z',
        'mtu': 0,
        'dns_domain': 'sample.openstack.org.',
        'vlan_transparent': None,
        'vlan_qinq': None,
        'segments': None,
    }

    network_availability_zone_extension = {
        "alias": "network_availability_zone",
        "updated": "2015-01-01T10:00:00-00:00",
        "description": "Availability zone support for router.",
        "links": [],
        "name": "Network Availability Zone",
    }

    enabled_neutron_extensions = [network_availability_zone_extension]

    def _compare_networks(self, exp, real):
        self.assertDictEqual(
            _network.Network(**exp).to_dict(computed=False),
            real.to_dict(computed=False),
        )

    def test_list_networks(self):
        net1 = {'id': '1', 'name': 'net1'}
        net2 = {'id': '2', 'name': 'net2'}
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'networks']
                    ),
                    json={'networks': [net1, net2]},
                )
            ]
        )
        nets = self.cloud.list_networks()
        self.assertEqual(
            [
                _network.Network(**i).to_dict(computed=False)
                for i in [net1, net2]
            ],
            [i.to_dict(computed=False) for i in nets],
        )
        self.assert_calls()

    def test_list_networks_filtered(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'networks'],
                        qs_elements=["name=test"],
                    ),
                    json={'networks': []},
                )
            ]
        )
        self.cloud.list_networks(filters={'name': 'test'})
        self.assert_calls()

    def test_list_networks_neutron_not_found(self):
        self.use_nothing()
        self.cloud.has_service = mock.Mock(return_value=False)
        self.assertEqual([], self.cloud.list_networks())
        self.assert_calls()

    def test_create_network(self):
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'networks']
                    ),
                    json={'network': self.mock_new_network_rep},
                    validate=dict(
                        json={
                            'network': {
                                'admin_state_up': True,
                                'name': 'netname',
                            }
                        }
                    ),
                )
            ]
        )
        network = self.cloud.create_network("netname")
        self._compare_networks(self.mock_new_network_rep, network)
        self.assert_calls()

    def test_create_network_specific_tenant(self):
        project_id = "project_id_value"
        mock_new_network_rep = copy.copy(self.mock_new_network_rep)
        mock_new_network_rep['project_id'] = project_id
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'networks']
                    ),
                    json={'network': mock_new_network_rep},
                    validate=dict(
                        json={
                            'network': {
                                'admin_state_up': True,
                                'name': 'netname',
                                'project_id': project_id,
                            }
                        }
                    ),
                )
            ]
        )
        network = self.cloud.create_network("netname", project_id=project_id)
        self._compare_networks(mock_new_network_rep, network)
        self.assert_calls()

    def test_create_network_external(self):
        mock_new_network_rep = copy.copy(self.mock_new_network_rep)
        mock_new_network_rep['router:external'] = True
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'networks']
                    ),
                    json={'network': mock_new_network_rep},
                    validate=dict(
                        json={
                            'network': {
                                'admin_state_up': True,
                                'name': 'netname',
                                'router:external': True,
                            }
                        }
                    ),
                )
            ]
        )
        network = self.cloud.create_network("netname", external=True)
        self._compare_networks(mock_new_network_rep, network)
        self.assert_calls()

    def test_create_network_provider(self):
        provider_opts = {
            'physical_network': 'mynet',
            'network_type': 'vlan',
            'segmentation_id': 'vlan1',
        }
        new_network_provider_opts = {
            'provider:physical_network': 'mynet',
            'provider:network_type': 'vlan',
            'provider:segmentation_id': 'vlan1',
        }
        mock_new_network_rep = copy.copy(self.mock_new_network_rep)
        mock_new_network_rep.update(new_network_provider_opts)
        expected_send_params = {'admin_state_up': True, 'name': 'netname'}
        expected_send_params.update(new_network_provider_opts)
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'networks']
                    ),
                    json={'network': mock_new_network_rep},
                    validate=dict(json={'network': expected_send_params}),
                )
            ]
        )
        network = self.cloud.create_network("netname", provider=provider_opts)
        self._compare_networks(mock_new_network_rep, network)
        self.assert_calls()

    def test_update_network_provider(self):
        network_id = "test-net-id"
        network_name = "network"
        network = {'id': network_id, 'name': network_name}
        provider_opts = {
            'physical_network': 'mynet',
            'network_type': 'vlan',
            'segmentation_id': 'vlan1',
            'should_not_be_passed': 1,
        }
        update_network_provider_opts = {
            'provider:physical_network': 'mynet',
            'provider:network_type': 'vlan',
            'provider:segmentation_id': 'vlan1',
        }
        mock_update_rep = copy.copy(self.mock_new_network_rep)
        mock_update_rep.update(update_network_provider_opts)
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'networks', network_name],
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'networks'],
                        qs_elements=[f'name={network_name}'],
                    ),
                    json={'networks': [network]},
                ),
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'networks', network_id],
                    ),
                    json={'network': mock_update_rep},
                    validate=dict(
                        json={'network': update_network_provider_opts}
                    ),
                ),
            ]
        )
        network = self.cloud.update_network(
            network_name, provider=provider_opts
        )
        self._compare_networks(mock_update_rep, network)
        self.assert_calls()

    def test_create_network_with_availability_zone_hints(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'extensions']
                    ),
                    json={'extensions': self.enabled_neutron_extensions},
                ),
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'networks']
                    ),
                    json={'network': self.mock_new_network_rep},
                    validate=dict(
                        json={
                            'network': {
                                'admin_state_up': True,
                                'name': 'netname',
                                'availability_zone_hints': ['nova'],
                            }
                        }
                    ),
                ),
            ]
        )
        network = self.cloud.create_network(
            "netname", availability_zone_hints=['nova']
        )
        self._compare_networks(self.mock_new_network_rep, network)
        self.assert_calls()

    def test_create_network_provider_ignored_value(self):
        provider_opts = {
            'physical_network': 'mynet',
            'network_type': 'vlan',
            'segmentation_id': 'vlan1',
            'should_not_be_passed': 1,
        }
        new_network_provider_opts = {
            'provider:physical_network': 'mynet',
            'provider:network_type': 'vlan',
            'provider:segmentation_id': 'vlan1',
        }
        mock_new_network_rep = copy.copy(self.mock_new_network_rep)
        mock_new_network_rep.update(new_network_provider_opts)
        expected_send_params = {'admin_state_up': True, 'name': 'netname'}
        expected_send_params.update(new_network_provider_opts)
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'networks']
                    ),
                    json={'network': mock_new_network_rep},
                    validate=dict(json={'network': expected_send_params}),
                )
            ]
        )
        network = self.cloud.create_network("netname", provider=provider_opts)
        self._compare_networks(mock_new_network_rep, network)
        self.assert_calls()

    def test_create_network_wrong_availability_zone_hints_type(self):
        azh_opts = "invalid"
        with testtools.ExpectedException(
            exceptions.SDKException,
            "Parameter 'availability_zone_hints' must be a list",
        ):
            self.cloud.create_network(
                "netname", availability_zone_hints=azh_opts
            )

    def test_create_network_provider_wrong_type(self):
        provider_opts = "invalid"
        with testtools.ExpectedException(
            exceptions.SDKException,
            "Parameter 'provider' must be a dict",
        ):
            self.cloud.create_network("netname", provider=provider_opts)

    def test_create_network_port_security_disabled(self):
        port_security_state = False
        mock_new_network_rep = copy.copy(self.mock_new_network_rep)
        mock_new_network_rep['port_security_enabled'] = port_security_state
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'networks']
                    ),
                    json={'network': mock_new_network_rep},
                    validate=dict(
                        json={
                            'network': {
                                'admin_state_up': True,
                                'name': 'netname',
                                'port_security_enabled': port_security_state,
                            }
                        }
                    ),
                )
            ]
        )
        network = self.cloud.create_network(
            "netname", port_security_enabled=port_security_state
        )
        self._compare_networks(mock_new_network_rep, network)
        self.assert_calls()

    def test_create_network_with_mtu(self):
        mtu_size = 1500
        mock_new_network_rep = copy.copy(self.mock_new_network_rep)
        mock_new_network_rep['mtu'] = mtu_size
        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'network', 'public', append=['v2.0', 'networks']
                    ),
                    json={'network': mock_new_network_rep},
                    validate=dict(
                        json={
                            'network': {
                                'admin_state_up': True,
                                'name': 'netname',
                                'mtu': mtu_size,
                            }
                        }
                    ),
                )
            ]
        )
        network = self.cloud.create_network("netname", mtu_size=mtu_size)
        self._compare_networks(mock_new_network_rep, network)
        self.assert_calls()

    def test_create_network_with_wrong_mtu_size(self):
        with testtools.ExpectedException(
            exceptions.SDKException,
            "Parameter 'mtu_size' must be greater than 67.",
        ):
            self.cloud.create_network("netname", mtu_size=42)

    def test_create_network_with_wrong_mtu_type(self):
        with testtools.ExpectedException(
            exceptions.SDKException,
            "Parameter 'mtu_size' must be an integer.",
        ):
            self.cloud.create_network("netname", mtu_size="fourty_two")

    def test_delete_network(self):
        network_id = "test-net-id"
        network_name = "network"
        network = {'id': network_id, 'name': network_name}
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'networks', network_name],
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'networks'],
                        qs_elements=[f'name={network_name}'],
                    ),
                    json={'networks': [network]},
                ),
                dict(
                    method='DELETE',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'networks', network_id],
                    ),
                    json={},
                ),
            ]
        )
        self.assertTrue(self.cloud.delete_network(network_name))
        self.assert_calls()

    def test_delete_network_not_found(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'networks', 'test-net'],
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'networks'],
                        qs_elements=['name=test-net'],
                    ),
                    json={'networks': []},
                ),
            ]
        )
        self.assertFalse(self.cloud.delete_network('test-net'))
        self.assert_calls()

    def test_delete_network_exception(self):
        network_id = "test-net-id"
        network_name = "network"
        network = {'id': network_id, 'name': network_name}
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'networks', network_name],
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'networks'],
                        qs_elements=[f'name={network_name}'],
                    ),
                    json={'networks': [network]},
                ),
                dict(
                    method='DELETE',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'networks', network_id],
                    ),
                    status_code=503,
                ),
            ]
        )
        self.assertRaises(
            exceptions.SDKException,
            self.cloud.delete_network,
            network_name,
        )
        self.assert_calls()

    def test_get_network_by_id(self):
        network_id = "test-net-id"
        network_name = "network"
        network = {'id': network_id, 'name': network_name}
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'network',
                        'public',
                        append=['v2.0', 'networks', network_id],
                    ),
                    json={'network': network},
                )
            ]
        )
        self.assertTrue(self.cloud.get_network_by_id(network_id))
        self.assert_calls()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from openstack import resource


class Attribute(resource.Resource):
    resource_key = 'attribute'
    resources_key = 'attributes'
    base_path = '/attributes'
    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = False
    allow_delete = True
    allow_list = True

    #: The timestamp when this attribute was created.
    created_at = resource.Body('created_at')
    #: The deployable_id of the attribute
    deployable_id = resource.Body('deployable_id')
    #: The key of the attribute
    key = resource.Body('key')
    #: The value of the attribute
    value = resource.Body('value')
    #: The timestamp when this attribute was updated.
    updated_at = resource.Body('updated_at')
    #: The uuid of the attribute
    uuid = resource.Body('uuid', alternate_id=True)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.baremetal.v1 import volume_connector
from openstack.tests.unit import base


FAKE = {
    "connector_id": "iqn.2017-07.org.openstack:01:d9a51732c3f",
    "created_at": "2016-08-18T22:28:48.643434+11:11",
    "extra": {},
    "links": [
        {
            "href": "http://127.0.0.1:6385/v1/volume/connector/<ID>",
            "rel": "self",
        },
        {
            "href": "http://127.0.0.1:6385/volume/connector/<ID>",
            "rel": "bookmark",
        },
    ],
    "node_uuid": "6d85703a-565d-469a-96ce-30b6de53079d",
    "type": "iqn",
    "updated_at": None,
    "uuid": "9bf93e01-d728-47a3-ad4b-5e66a835037c",
}


class TestVolumeconnector(base.TestCase):
    def test_basic(self):
        sot = volume_connector.VolumeConnector()
        self.assertIsNone(sot.resource_key)
        self.assertEqual('connectors', sot.resources_key)
        self.assertEqual('/volume/connectors', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)
        self.assertEqual('PATCH', sot.commit_method)

    def test_instantiate(self):
        sot = volume_connector.VolumeConnector(**FAKE)
        self.assertEqual(FAKE['connector_id'], sot.connector_id)
        self.assertEqual(FAKE['created_at'], sot.created_at)
        self.assertEqual(FAKE['extra'], sot.extra)
        self.assertEqual(FAKE['links'], sot.links)
        self.assertEqual(FAKE['node_uuid'], sot.node_id)
        self.assertEqual(FAKE['type'], sot.type)
        self.assertEqual(FAKE['updated_at'], sot.updated_at)
        self.assertEqual(FAKE['uuid'], sot.id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class TapService(resource.Resource):
    """Tap Service"""

    resource_key = 'tap_service'
    resources_key = 'tap_services'
    base_path = '/taas/tap_services'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    _allow_unknown_attrs_in_body = True

    _query_mapping = resource.QueryParameters(
        "sort_key", "sort_dir", 'name', 'project_id'
    )

    # Properties
    #: The ID of the tap service.
    id = resource.Body('id')
    #: The tap service name.
    name = resource.Body('name')
    #: The tap service description.
    description = resource.Body('description')
    #: The ID of the project that owns the tap service.
    project_id = resource.Body('project_id', alias='tenant_id')
    #: Tenant_id (deprecated attribute).
    tenant_id = resource.Body('tenant_id', deprecated=True)
    #: The id of the port the tap service is associated with
    port_id = resource.Body('port_id')
    #: The status for the tap service.
    status = resource.Body('status')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from openstack import exceptions
from openstack import resource
from openstack import utils


class Service(resource.Resource):
    resources_key = 'services'
    base_path = '/os-services'

    # capabilities
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'binary',
        'host',
    )

    # Properties
    #: The ID of active storage backend (cinder-volume services only)
    active_backend_id = resource.Body('active_backend_id')
    #: The availability zone of service
    availability_zone = resource.Body('zone')
    #: Binary name of service
    binary = resource.Body('binary')
    #: Disabled reason of service
    disabled_reason = resource.Body('disabled_reason')
    #: The name of the host where service runs
    host = resource.Body('host')
    # Whether the host is frozen or not (cinder-volume services only)
    is_frozen = resource.Body('frozen')
    #: Service name
    name = resource.Body('name', alias='binary')
    #: The volume service replication status (cinder-volume services only)
    replication_status = resource.Body('replication_status')
    #: State of service
    state = resource.Body('state')
    #: Status of service
    status = resource.Body('status')
    #: The date and time when the resource was updated
    updated_at = resource.Body('updated_at')

    @classmethod
    def find(cls, session, name_or_id, ignore_missing=True, **params):
        # No direct request possible, thus go directly to list
        data = cls.list(session, **params)

        result = None
        for maybe_result in data:
            # Since ID might be both int and str force cast
            id_value = str(cls._get_id(maybe_result))
            name_value = maybe_result.name

            if str(name_or_id) in (id_value, name_value):
                if 'host' in params and maybe_result['host'] != params['host']:
                    continue
                # Only allow one resource to be found. If we already
                # found a match, raise an exception to show it.
                if result is None:
                    result = maybe_result
                else:
                    msg = "More than one %s exists with the name '%s'."
                    msg = msg % (cls.__name__, name_or_id)
                    raise exceptions.DuplicateResource(msg)

        if result is not None:
            return result

        if ignore_missing:
            return None
        raise exceptions.NotFoundException(
            f"No {cls.__name__} found for {name_or_id}"
        )

    def commit(self, session, prepend_key=False, *args, **kwargs):
        # we need to set prepend_key to false
        return super().commit(
            session,
            prepend_key,
            *args,
            **kwargs,
        )

    def _action(self, session, action, body):
        url = utils.urljoin(Service.base_path, action)
        response = session.put(url, json=body)
        self._translate_response(response)
        return self

    def enable(self, session):
        """Enable service."""
        body = {'binary': self.binary, 'host': self.host}
        return self._action(session, 'enable', body)

    def disable(self, session, *, reason=None):
        """Disable service."""
        body = {'binary': self.binary, 'host': self.host}

        if not reason:
            action = 'disable'
        else:
            action = 'disable-log-reason'
            body['disabled_reason'] = reason

        return self._action(session, action, body)

    def thaw(self, session):
        body = {'host': self.host}
        return self._action(session, 'thaw', body)

    def freeze(self, session):
        body = {'host': self.host}
        return self._action(session, 'freeze', body)

    def failover(
        self,
        session,
        *,
        backend_id=None,
    ):
        """Failover a service

        Only applies to replicating cinder-volume services.
        """
        body = {'host': self.host}
        if backend_id:
            body['backend_id'] = backend_id

        return self._action(session, 'failover_host', body)



# Copyright(c) 2018 Nippon Telegraph and Telephone Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from openstack import resource


class Host(resource.Resource):
    resource_key = "host"
    resources_key = "hosts"
    base_path = "/segments/%(segment_id)s/hosts"

    # capabilities
    # 1] GET /v1/segments/<segment_uuid>/hosts
    # 2] GET /v1/segments/<segment_uuid>/hosts/<host_uuid>
    # 3] POST /v1/segments/<segment_uuid>/hosts
    # 4] PUT /v1/segments/<segment_uuid>/hosts
    # 5] DELETE /v1/segments/<segment_uuid>/hosts
    allow_list = True
    allow_fetch = True
    allow_create = True
    allow_commit = True
    allow_delete = True

    #: A Uuid of representing this host
    uuid = resource.Body("uuid")
    #: A failover segment ID of this host(in URI)
    segment_id = resource.URI("segment_id")
    #: A created time of this host
    created_at = resource.Body("created_at")
    #: A latest updated time of this host
    updated_at = resource.Body("updated_at")
    #: A name of this host
    name = resource.Body("name")
    #: A type of this host
    type = resource.Body("type")
    #: A control attributes of this host
    control_attributes = resource.Body("control_attributes")
    #: A maintenance status of this host
    on_maintenance = resource.Body("on_maintenance")
    #: A reservation status of this host
    reserved = resource.Body("reserved")
    #: A failover segment ID of this host(in Body)
    failover_segment_id = resource.Body("failover_segment_id")

    _query_mapping = resource.QueryParameters(
        "sort_key",
        "sort_dir",
        failover_segment_id="failover_segment_id",
        type="type",
        on_maintenance="on_maintenance",
        reserved="reserved",
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from openstack import resource


class Event(resource.Resource):
    resource_key = 'event'
    resources_key = 'events'
    base_path = '/events'

    # Capabilities
    allow_list = True
    allow_fetch = True

    _query_mapping = resource.QueryParameters(
        'cluster_id',
        'action',
        'level',
        'sort',
        'global_project',
        obj_id='oid',
        obj_name='oname',
        obj_type='otype',
    )

    # Properties
    #: Timestamp string (in ISO8601 format) when the event was generated.
    generated_at = resource.Body('timestamp')
    #: The UUID of the object related to this event.
    obj_id = resource.Body('oid')
    #: The name of the object related to this event.
    obj_name = resource.Body('oname')
    #: The type name of the object related to this event.
    obj_type = resource.Body('otype')
    #: The UUID of the cluster related to this event, if any.
    cluster_id = resource.Body('cluster_id')
    #: The event level (priority).
    level = resource.Body('level')
    #: The ID of the user.
    user_id = resource.Body('user')
    #: The ID of the project (tenant).
    project_id = resource.Body('project')
    #: The string representation of the action associated with the event.
    action = resource.Body('action')
    #: The status of the associated object.
    status = resource.Body('status')
    #: A string description of the reason that brought the object into its
    #: current status.
    status_reason = resource.Body('status_reason')
    #: The metadata of an event object.
    meta_data = resource.Body('meta_data')



# Licensed under the Apache License, Version 2.0 (the "License"); you may

# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

from openstack import exceptions
from openstack import resource


class BaseResource(resource.Resource):
    commit_method = 'POST'
    create_method = 'PUT'

    #: Metadata stored for this resource. *Type: dict*
    metadata: dict[str, ty.Any] = {}

    _custom_metadata_prefix: str
    _system_metadata: dict[str, ty.Any] = {}
    _last_headers: dict[str, ty.Any] = {}

    def __init__(self, metadata=None, **attrs):
        """Process and save metadata known at creation stage"""
        super().__init__(**attrs)
        if metadata is not None:
            for k, v in metadata.items():
                if not k.lower().startswith(
                    self._custom_metadata_prefix.lower()
                ):
                    self.metadata[self._custom_metadata_prefix + k] = v
                else:
                    self.metadata[k] = v

    def _prepare_request(self, *args, **kwargs):
        request = super()._prepare_request(*args, **kwargs)
        request.headers.update(self._calculate_headers(self.metadata))
        return request

    def _calculate_headers(self, metadata):
        headers = {}
        for key in metadata:
            if key in self._system_metadata.keys():
                header = self._system_metadata[key]
            elif key in self._system_metadata.values():
                header = key
            else:
                if key.startswith(self._custom_metadata_prefix):
                    header = key
                else:
                    header = self._custom_metadata_prefix + key
            headers[header] = str(metadata[key])
        return headers

    def set_metadata(self, session, metadata, refresh=True):
        request = self._prepare_request()
        response = session.post(
            request.url, headers=self._calculate_headers(metadata)
        )
        self._translate_response(response, has_body=False)
        if refresh:
            response = session.head(request.url)
            self._translate_response(response, has_body=False)
        return self

    def delete_metadata(self, session, keys):
        request = self._prepare_request()
        headers = {key: '' for key in keys}
        response = session.post(
            request.url, headers=self._calculate_headers(headers)
        )
        exceptions.raise_from_response(
            response, error_message="Error deleting metadata keys"
        )
        return self

    def _set_metadata(self, headers):
        self.metadata = dict()

        for header in headers:
            # RADOS and other stuff in front may actually lowcase headers
            if header.lower().startswith(self._custom_metadata_prefix.lower()):
                key = header[len(self._custom_metadata_prefix) :].lower()
                self.metadata[key] = headers[header]

    def _translate_response(
        self,
        response,
        has_body=None,
        error_message=None,
        *,
        resource_response_key=None,
    ):
        # Save headers of the last operation for potential use (get_object of
        # cloud layer).
        # This must happen before invoking parent _translate_response, cause it
        # pops known headers.
        self._last_headers = response.headers.copy()
        super()._translate_response(
            response, has_body=has_body, error_message=error_message
        )
        self._set_metadata(response.headers)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from openstack import resource


class Claim(resource.Resource):
    # FIXME(anyone): The name string of `location` field of Zaqar API response
    # is lower case. That is inconsistent with the guide from API-WG. This is
    # a workaround for this issue.
    location = resource.Header("location")

    resources_key = 'claims'
    base_path = '/queues/%(queue_name)s/claims'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    commit_method = 'PATCH'

    # Properties
    #: The value in seconds indicating how long the claim has existed.
    age = resource.Body("age")
    #: In case worker stops responding for a long time, the server will
    #: extend the lifetime of claimed messages to be at least as long as
    #: the lifetime of the claim itself, plus the specified grace period.
    #: Must between 60 and 43200 seconds(12 hours).
    grace = resource.Body("grace")
    #: The number of messages to claim. Default 10, up to 20.
    limit = resource.Body("limit")
    #: Messages have been successfully claimed.
    messages = resource.Body("messages")
    #: Number of seconds the server wait before releasing the claim. Must
    #: between 60 and 43200 seconds(12 hours).
    ttl = resource.Body("ttl")
    #: The name of queue to claim message from.
    queue_name = resource.URI("queue_name")
    #: The ID to identify the client accessing Zaqar API. Must be specified
    #: in header for each API request.
    client_id = resource.Header("Client-ID")
    #: The ID to identify the project. Must be provided when keystone
    #: authentication is not enabled in Zaqar service.
    project_id = resource.Header("X-PROJECT-ID")

    def _translate_response(
        self,
        response,
        has_body=None,
        error_message=None,
        *,
        resource_response_key=None,
    ):
        super()._translate_response(
            response,
            has_body,
            error_message,
            resource_response_key=resource_response_key,
        )
        if has_body and self.location:
            # Extract claim ID from location
            self.id = self.location.split("claims/")[1]

    def create(self, session, prepend_key=False, base_path=None, **kwargs):
        request = self._prepare_request(
            requires_id=False, prepend_key=prepend_key, base_path=base_path
        )
        headers = {
            "Client-ID": self.client_id or str(uuid.uuid4()),
            "X-PROJECT-ID": self.project_id or session.get_project_id(),
        }
        request.headers.update(headers)
        response = session.post(
            request.url, json=request.body, headers=request.headers
        )

        # For case no message was claimed successfully, 204 No Content
        # message will be returned. In other cases, we translate response
        # body which has `messages` field(list) included.
        if response.status_code != 204:
            self._translate_response(response)

        return self

    def fetch(
        self,
        session,
        requires_id=True,
        base_path=None,
        error_message=None,
        skip_cache=False,
        **kwargs,
    ):
        request = self._prepare_request(
            requires_id=requires_id, base_path=base_path
        )
        headers = {
            "Client-ID": self.client_id or str(uuid.uuid4()),
            "X-PROJECT-ID": self.project_id or session.get_project_id(),
        }

        request.headers.update(headers)
        response = session.get(
            request.url, headers=request.headers, skip_cache=False
        )
        self._translate_response(response)

        return self

    def commit(
        self,
        session,
        prepend_key=True,
        has_body=True,
        retry_on_conflict=None,
        base_path=None,
        **kwargs,
    ):
        request = self._prepare_request(
            prepend_key=prepend_key, base_path=base_path
        )
        headers = {
            "Client-ID": self.client_id or str(uuid.uuid4()),
            "X-PROJECT-ID": self.project_id or session.get_project_id(),
        }

        request.headers.update(headers)
        session.patch(request.url, json=request.body, headers=request.headers)

        return self

    def delete(self, session, *args, **kwargs):
        request = self._prepare_request()
        headers = {
            "Client-ID": self.client_id or str(uuid.uuid4()),
            "X-PROJECT-ID": self.project_id or session.get_project_id(),
        }

        request.headers.update(headers)
        response = session.delete(request.url, headers=request.headers)

        self._translate_response(response, has_body=False)
        return self



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import exceptions
from openstack import resource
from openstack import utils


class Aggregate(resource.Resource):
    resource_key = 'aggregate'
    resources_key = 'aggregates'
    base_path = '/os-aggregates'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_delete = True
    allow_list = True
    allow_commit = True

    # Properties
    #: Availability zone of aggregate
    availability_zone = resource.Body('availability_zone')
    #: The date and time when the resource was created.
    created_at = resource.Body('created_at')
    #: The date and time when the resource was deleted.
    deleted_at = resource.Body('deleted_at')
    #: Deleted?
    is_deleted = resource.Body('deleted', type=bool)
    #: Name of aggregate
    name = resource.Body('name')
    #: Hosts
    hosts = resource.Body('hosts', type=list)
    #: Metadata
    metadata = resource.Body('metadata', type=dict)
    #: The date and time when the resource was updated
    updated_at = resource.Body('updated_at')
    #: UUID
    uuid = resource.Body('uuid')
    # Image pre-caching introduced in 2.81
    _max_microversion = '2.81'

    def _action(self, session, body, microversion=None):
        """Preform aggregate actions given the message body."""
        url = utils.urljoin(self.base_path, self.id, 'action')
        response = session.post(url, json=body, microversion=microversion)
        exceptions.raise_from_response(response)
        aggregate = Aggregate()
        aggregate._translate_response(response)
        return aggregate

    def add_host(self, session, host):
        """Adds a host to an aggregate."""
        body = {'add_host': {'host': host}}
        return self._action(session, body)

    def remove_host(self, session, host):
        """Removes a host from an aggregate."""
        body = {'remove_host': {'host': host}}
        return self._action(session, body)

    def set_metadata(self, session, metadata):
        """Creates or replaces metadata for an aggregate."""
        body = {'set_metadata': {'metadata': metadata}}
        return self._action(session, body)

    def precache_images(self, session, images):
        """Requests image pre-caching"""
        body = {'cache': images}
        url = utils.urljoin(self.base_path, self.id, 'images')
        response = session.post(
            url, json=body, microversion=self._max_microversion
        )
        exceptions.raise_from_response(response)
        # This API has no result



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from openstack.load_balancer.v2 import availability_zone
from openstack.tests.unit import base


AVAILABILITY_ZONE_PROFILE_ID = uuid.uuid4()
EXAMPLE = {
    'name': 'strawberry',
    'description': 'tasty',
    'is_enabled': False,
    'availability_zone_profile_id': AVAILABILITY_ZONE_PROFILE_ID,
}


class TestAvailabilityZone(base.TestCase):
    def test_basic(self):
        test_availability_zone = availability_zone.AvailabilityZone()
        self.assertEqual(
            'availability_zone', test_availability_zone.resource_key
        )
        self.assertEqual(
            'availability_zones', test_availability_zone.resources_key
        )
        self.assertEqual(
            '/lbaas/availabilityzones', test_availability_zone.base_path
        )
        self.assertTrue(test_availability_zone.allow_create)
        self.assertTrue(test_availability_zone.allow_fetch)
        self.assertTrue(test_availability_zone.allow_commit)
        self.assertTrue(test_availability_zone.allow_delete)
        self.assertTrue(test_availability_zone.allow_list)

    def test_make_it(self):
        test_availability_zone = availability_zone.AvailabilityZone(**EXAMPLE)
        self.assertEqual(EXAMPLE['name'], test_availability_zone.name)
        self.assertEqual(
            EXAMPLE['description'], test_availability_zone.description
        )
        self.assertFalse(test_availability_zone.is_enabled)
        self.assertEqual(
            EXAMPLE['availability_zone_profile_id'],
            test_availability_zone.availability_zone_profile_id,
        )

        self.assertDictEqual(
            {
                'limit': 'limit',
                'marker': 'marker',
                'name': 'name',
                'description': 'description',
                'is_enabled': 'enabled',
                'availability_zone_profile_id': 'availability_zone_profile_id',
            },
            test_availability_zone._query_mapping._mapping,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Create resources with the Network service.

For a full guide see
https://docs.openstack.org/openstacksdk/latest/user/guides/network.html
"""


def create_network(conn):
    print("Create Network:")

    example_network = conn.network.create_network(
        name='openstacksdk-example-project-network'
    )

    print(example_network)

    example_subnet = conn.network.create_subnet(
        name='openstacksdk-example-project-subnet',
        network_id=example_network.id,
        ip_version='4',
        cidr='10.0.2.0/24',
        gateway_ip='10.0.2.1',
    )

    print(example_subnet)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests import fakes
from openstack.tests.unit import base


class TestAggregate(base.TestCase):
    def setUp(self):
        super().setUp()
        self.aggregate_name = self.getUniqueString('aggregate')
        self.fake_aggregate = fakes.make_fake_aggregate(1, self.aggregate_name)
        self.use_compute_discovery()

    def test_create_aggregate(self):
        create_aggregate = self.fake_aggregate.copy()
        del create_aggregate['metadata']
        del create_aggregate['hosts']

        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['os-aggregates']
                    ),
                    json={'aggregate': create_aggregate},
                    validate=dict(
                        json={
                            'aggregate': {
                                'name': self.aggregate_name,
                                'availability_zone': None,
                            }
                        }
                    ),
                ),
            ]
        )
        self.cloud.create_aggregate(name=self.aggregate_name)

        self.assert_calls()

    def test_create_aggregate_with_az(self):
        availability_zone = 'az1'
        az_aggregate = fakes.make_fake_aggregate(
            1, self.aggregate_name, availability_zone=availability_zone
        )

        create_aggregate = az_aggregate.copy()
        del create_aggregate['metadata']
        del create_aggregate['hosts']

        self.register_uris(
            [
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['os-aggregates']
                    ),
                    json={'aggregate': create_aggregate},
                    validate=dict(
                        json={
                            'aggregate': {
                                'name': self.aggregate_name,
                                'availability_zone': availability_zone,
                            }
                        }
                    ),
                ),
            ]
        )

        self.cloud.create_aggregate(
            name=self.aggregate_name, availability_zone=availability_zone
        )

        self.assert_calls()

    def test_delete_aggregate(self):
        self.register_uris(
            [
                dict(
                    method='DELETE',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['os-aggregates', '1']
                    ),
                ),
            ]
        )

        self.assertTrue(self.cloud.delete_aggregate('1'))

        self.assert_calls()

    def test_delete_aggregate_by_name(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['os-aggregates', self.aggregate_name],
                    ),
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['os-aggregates']
                    ),
                    json={'aggregates': [self.fake_aggregate]},
                ),
                dict(
                    method='DELETE',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['os-aggregates', '1']
                    ),
                ),
            ]
        )

        self.assertTrue(self.cloud.delete_aggregate(self.aggregate_name))

        self.assert_calls()

    def test_update_aggregate_set_az(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['os-aggregates', '1']
                    ),
                    json=self.fake_aggregate,
                ),
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['os-aggregates', '1']
                    ),
                    json={'aggregate': self.fake_aggregate},
                    validate=dict(
                        json={
                            'aggregate': {
                                'availability_zone': 'az',
                            }
                        }
                    ),
                ),
            ]
        )

        self.cloud.update_aggregate(1, availability_zone='az')

        self.assert_calls()

    def test_update_aggregate_unset_az(self):
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['os-aggregates', '1']
                    ),
                    json=self.fake_aggregate,
                ),
                dict(
                    method='PUT',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['os-aggregates', '1']
                    ),
                    json={'aggregate': self.fake_aggregate},
                    validate=dict(
                        json={
                            'aggregate': {
                                'availability_zone': None,
                            }
                        }
                    ),
                ),
            ]
        )

        self.cloud.update_aggregate(1, availability_zone=None)

        self.assert_calls()

    def test_set_aggregate_metadata(self):
        metadata = {'key': 'value'}
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['os-aggregates', '1']
                    ),
                    json=self.fake_aggregate,
                ),
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['os-aggregates', '1', 'action'],
                    ),
                    json={'aggregate': self.fake_aggregate},
                    validate=dict(
                        json={'set_metadata': {'metadata': metadata}}
                    ),
                ),
            ]
        )
        self.cloud.set_aggregate_metadata('1', metadata)

        self.assert_calls()

    def test_add_host_to_aggregate(self):
        hostname = 'host1'
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['os-aggregates', '1']
                    ),
                    json=self.fake_aggregate,
                ),
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['os-aggregates', '1', 'action'],
                    ),
                    json={'aggregate': self.fake_aggregate},
                    validate=dict(json={'add_host': {'host': hostname}}),
                ),
            ]
        )
        self.cloud.add_host_to_aggregate('1', hostname)

        self.assert_calls()

    def test_remove_host_from_aggregate(self):
        hostname = 'host1'
        self.register_uris(
            [
                dict(
                    method='GET',
                    uri=self.get_mock_url(
                        'compute', 'public', append=['os-aggregates', '1']
                    ),
                    json=self.fake_aggregate,
                ),
                dict(
                    method='POST',
                    uri=self.get_mock_url(
                        'compute',
                        'public',
                        append=['os-aggregates', '1', 'action'],
                    ),
                    json={'aggregate': self.fake_aggregate},
                    validate=dict(json={'remove_host': {'host': hostname}}),
                ),
            ]
        )
        self.cloud.remove_host_from_aggregate('1', hostname)

        self.assert_calls()



# Copyright (C) 2018 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from openstack.compute.v2 import hypervisor
from openstack import connection
from openstack.tests.functional import base

HYPERVISORS: list[hypervisor.Hypervisor] = []


def hypervisors():
    global HYPERVISORS
    if HYPERVISORS:
        return True
    HYPERVISORS = connection.Connection.list_hypervisors(
        connection.from_config(cloud_name=base.TEST_CLOUD_NAME)
    )
    return bool(HYPERVISORS)


class TestHost(base.BaseFunctionalTest):
    def setUp(self):
        super().setUp()
        self.require_service('instance-ha')
        self.NAME = self.getUniqueString()

        if not hypervisors():
            self.skipTest(
                "Skip TestHost as there are no hypervisors configured in nova"
            )

        # Create segment
        self.segment = self.operator_cloud.ha.create_segment(
            name=self.NAME, recovery_method='auto', service_type='COMPUTE'
        )

        # Create valid host
        self.NAME = HYPERVISORS[0].name
        self.host = self.operator_cloud.ha.create_host(
            segment_id=self.segment.uuid,
            name=self.NAME,
            type='COMPUTE',
            control_attributes='SSH',
        )

        # Delete host
        self.addCleanup(
            self.operator_cloud.ha.delete_host,
            self.segment.uuid,
            self.host.uuid,
        )
        # Delete segment
        self.addCleanup(
            self.operator_cloud.ha.delete_segment, self.segment.uuid
        )

    def test_list(self):
        names = [
            o.name
            for o in self.operator_cloud.ha.hosts(
                self.segment.uuid,
                failover_segment_id=self.segment.uuid,
                type='COMPUTE',
            )
        ]
        self.assertIn(self.NAME, names)

    def test_update(self):
        updated_host = self.operator_cloud.ha.update_host(
            self.host['uuid'],
            segment_id=self.segment.uuid,
            on_maintenance='True',
        )
        get_host = self.operator_cloud.ha.get_host(
            updated_host.uuid, updated_host.segment_id
        )
        self.assertEqual(True, get_host.on_maintenance)



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import resource


class AvailabilityZoneProfile(resource.Resource):
    resource_key = 'availability_zone_profile'
    resources_key = 'availability_zone_profiles'
    base_path = '/lbaas/availabilityzoneprofiles'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    _query_mapping = resource.QueryParameters(
        'id', 'name', 'provider_name', 'availability_zone_data'
    )

    # Properties
    #: The ID of the availability zone profile.
    id = resource.Body('id')
    #: The name of the availability zone profile.
    name = resource.Body('name')
    #: The provider this availability zone profile is for.
    provider_name = resource.Body('provider_name')
    #: The JSON string containing the availability zone metadata.
    availability_zone_data = resource.Body('availability_zone_data')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Helpers for building configdrive compatible with the Bare Metal service."""

import base64
import contextlib
import gzip
import json
import os
import shutil
import subprocess
import tempfile


@contextlib.contextmanager
def populate_directory(
    metadata,
    user_data=None,
    versions=None,
    network_data=None,
    vendor_data=None,
):
    """Populate a directory with configdrive files.

    :param dict metadata: Metadata.
    :param bytes user_data: Vendor-specific user data.
    :param versions: List of metadata versions to support.
    :param dict network_data: Networking configuration.
    :param dict vendor_data: Extra supplied vendor data.
    :return: a context manager yielding a directory with files
    """
    d = tempfile.mkdtemp()
    versions = versions or ('2012-08-10', 'latest')
    try:
        for version in versions:
            subdir = os.path.join(d, 'openstack', version)
            if not os.path.exists(subdir):
                os.makedirs(subdir)

            with open(os.path.join(subdir, 'meta_data.json'), 'w') as fp:
                json.dump(metadata, fp)

            if network_data:
                with open(
                    os.path.join(subdir, 'network_data.json'), 'w'
                ) as fp:
                    json.dump(network_data, fp)

            if vendor_data:
                with open(
                    os.path.join(subdir, 'vendor_data2.json'), 'w'
                ) as fp:
                    json.dump(vendor_data, fp)

            if user_data:
                # Strictly speaking, user data is binary, but in many cases
                # it's actually a text (cloud-init, ignition, etc).
                flag = 't' if isinstance(user_data, str) else 'b'
                with open(os.path.join(subdir, 'user_data'), f'w{flag}') as fp:
                    fp.write(user_data)

        yield d
    finally:
        shutil.rmtree(d)


def build(
    metadata,
    user_data=None,
    versions=None,
    network_data=None,
    vendor_data=None,
):
    """Make a configdrive compatible with the Bare Metal service.

    Requires the genisoimage utility to be available.

    :param dict metadata: Metadata.
    :param user_data: Vendor-specific user data.
    :param versions: List of metadata versions to support.
    :param dict network_data: Networking configuration.
    :param dict vendor_data: Extra supplied vendor data.
    :return: configdrive contents as a base64-encoded string.
    """
    with populate_directory(
        metadata, user_data, versions, network_data, vendor_data
    ) as path:
        return pack(path)


def pack(path: str) -> str:
    """Pack a directory with files into a Bare Metal service configdrive.

    Creates an ISO image with the files and label "config-2".

    :param str path: Path to directory with files
    :return: configdrive contents as a base64-encoded string.
    """
    with tempfile.NamedTemporaryFile() as tmpfile:
        # NOTE(toabctl): Luckily, genisoimage, mkisofs and xorrisofs understand
        # the same parameters which are currently used.
        error: Exception | None
        for c in ['genisoimage', 'mkisofs', 'xorrisofs']:
            try:
                p = subprocess.Popen(  # noqa: S603
                    [
                        c,
                        '-o',
                        tmpfile.name,
                        '-ldots',
                        '-allow-lowercase',
                        '-allow-multidot',
                        '-l',
                        '-publisher',
                        'metalsmith',
                        '-quiet',
                        '-J',
                        '-r',
                        '-V',
                        'config-2',
                        path,
                    ],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
            except OSError as e:
                error = e
            else:
                error = None
                break

        if error:
            raise RuntimeError(
                'Error generating the configdrive. Make sure the '
                '"genisoimage", "mkisofs" or "xorrisofs" tool is installed. '
                f'Error: {error}'
            )

        stdout, stderr = p.communicate()
        if p.returncode != 0:
            raise RuntimeError(
                'Error generating the configdrive.'
                f'Stdout: "{stdout.decode()}". Stderr: "{stderr.decode()}"'
            )

        tmpfile.seek(0)

        with tempfile.NamedTemporaryFile() as tmpzipfile:
            with gzip.GzipFile(fileobj=tmpzipfile, mode='wb') as gz_file:
                shutil.copyfileobj(tmpfile, gz_file)

            tmpzipfile.seek(0)
            # NOTE(dtantsur): Ironic expects configdrive to be a string, but
            # base64 returns bytes on Python 3.
            cd = base64.b64encode(tmpzipfile.read()).decode()

    return cd
